import logging
import sys
from lusid.utilities import ApiClientFactory

from lusidtools.apps.upsert_instruments import load_mapping_file_for_file_type
from lusidtools.logger import LusidLogger
from lusidtools.cocoon import parse_args, load_data_to_df_and_detect_delimiter, load_from_data_frame, \
    identify_cash_items, validate_mapping_file_structure


def load_holdings(args):
    file_type = "holdings"

    factory = ApiClientFactory(api_secrets_filename=args["secrets_file"])

    if args['delimiter']:
        logging.info(f"delimiter specified as {repr(args['delimiter'])}")
    logging.debug("Getting data")
    holdings = load_data_to_df_and_detect_delimiter(args)

    mappings = load_mapping_file_for_file_type(args["mapping"], file_type)
    if "cash_flag" in mappings.keys():
        holdings, mappings = identify_cash_items(holdings, mappings)

    validate_mapping_file_structure(mappings, holdings.columns, file_type)

    if args["dryrun"]:
        logging.info("--dryrun specified as True, exiting before upsert call is made")
        return 0

    holdings_response = load_from_data_frame(api_factory=factory,
                                             data_frame=holdings,
                                             scope=args["scope"],
                                             identifier_mapping=mappings["identifier_mapping"],
                                             mapping_required=mappings["required"],
                                             mapping_optional=mappings["optional"],
                                             file_type=file_type,
                                             batch_size=args["batch_size"],
                                             property_columns=mappings[
                                                 "property_columns"] if "property_columns" in mappings.keys() else [])

    total_success = len(holdings_response['holdings']['success'])
    total_failed = len(holdings_response['holdings']['errors'])
    logging.info(f"Success: {total_success}/{total_success + total_failed}")
    logging.info(f"Fail:    {total_failed}/{total_success + total_failed}")
    return holdings_response


def main(argv):
    args, ap = parse_args(sys.argv[1:])
    LusidLogger(args["debug"])
    load_holdings(args)

    return 0


if __name__ == "__main__":
    main(sys.argv)



import logging


class Validator:

    def __init__(self, value, value_name):
        self.value = value
        self.value_name = value_name

    def check_allowed_value(self, allowed_values: list):
        """
        Checks that value exists in the provided list

        :param list allowed_values: The list of allowed values

        :return: self
        """
        if self.value not in allowed_values:
            raise ValueError(f"The {self.value_name} must be one of {str(allowed_values)}, you supplied '{self.value}' instead.")
        return self

    def make_singular(self):
        """
        Makes a plural string singular

        :return: self
        """
        if isinstance(self.value, str):
            self.value = self.value.rstrip("s")
            logging.debug(f"The value of {self.value_name} has had any 's' characters stripped from the right to make it singular")
        return self

    def make_lower(self):
        """
        Makes a string lowercase

        :return: self
        """
        if isinstance(self.value, str):
            self.value = self.value.lower()
            logging.debug(
            f"The value of {self.value_name} has been made lower case")
        return self

    def set_default_value_if_none(self, default):
        """
        Sets a default value if the current value is None

        :param default: The default value

        :return: self
        """
        if self.value is None:
            self.value = default
            logging.debug(f"The value of {self.value_name} has been updated from None to {default}")
        return self

    def override_value(self, override_flag: bool, override_value):
        """
        Overrides the current value of the ovverride_flag is True

        :param bool override_flag: Whether or not to override the current value
        :param override_value: The value to use to override the existing value

        :return: self
        """
        if override_flag:
            self.value = override_value
            logging.debug(f"The value of {self.value_name} has been overriden with {override_value}")
        return self

    def discard_dict_keys_none_value(self):
        """
        Discards dictionary key, value pairs where the value is None

        :return: self
        """
        if isinstance(self.value, dict):
            invalid_keys = [key for key, value in self.value.items() if value is None]
            if len(invalid_keys) > 0:
                logging.info(f"The values for the keys {str(invalid_keys)} are None and have thus been removed from {self.value_name}")
                for key in invalid_keys:
                    self.value.pop(key, None)
        return self

    def get_dict_values(self):
        """
        Gets a list of values from a dictionary

        :return: self
        """
        if isinstance(self.value, dict):
            self.value = list(self.value.values())
        return self

    def filter_list_using_first_character(self, first_character: str):
        """
        Filters a list of strings by looking to see if the first character matches the provided

        :param str first_character: The character to look for in the first character of each element
        :return: self
        """
        if isinstance(self.value, list):
            to_remove = []
            for value in self.value:
                if isinstance(value, str) and value[0] == first_character:
                    to_remove.append(value)
            self.value = [value for value in self.value if value not in to_remove]
        return self

    def check_subset_of_list(self, superset: list, superset_name: str):
        """
        Checks if one list is a subset of another

        :param list superset: The superset to check of the value is a subset of
        :param list superset_name: The name of the superset

        :return: self
        """
        if isinstance(self.value, list):
            if len(set(self.value) - set(superset)) > 0:
                raise ValueError(f'''The values {str(set(self.value) - set(superset))} exist in the {self.value_name}
                                   but do not exist in the {superset_name}.''')
        return self

    def check_no_intersection_with_list(self, other_list: list, list_name: str):
        """
        Checks that the value has no intersection with a provided list

        :param list other_list: The list to check the value has no intersection with
        :param str list_name: The name of the list

        :return: self
        """
        if isinstance(self.value, list):
            if len(set(other_list).intersection(set(self.value))) > 0:
                raise ValueError(f'''The columns {str(set(other_list).intersection(set(self.value)))} are specified in {self.value_name}
                                     yet they contain null (NaN) values for some rows in the provided data. Null values are not 
                                     allowed for required fields. Please ensure that required columns do not contain ANY null 
                                     values or specify a default value in your mapping by specifying a dictionary with the keys
                                     "column" and "default".''')



import lusidtools.cocoon.cocoon
import lusidtools.cocoon.instruments
import lusidtools.cocoon.properties
import lusidtools.cocoon.systemConfiguration
import lusidtools.cocoon.utilities
from lusidtools.cocoon.instruments import resolve_instruments
from lusidtools.cocoon.properties import create_property_values
from lusidtools.cocoon.utilities import set_attributes
from lusidtools.cocoon.cocoon import load_from_data_frame
from lusidtools.cocoon.utilities import checkargs, load_data_to_df_and_detect_delimiter, check_mapping_fields_exist, parse_args, identify_cash_items, validate_mapping_file_structure


import lusidtools.cocoon.async_tools
import lusidtools.cocoon.validator
import lusidtools.cocoon.dateorcutlabel


import lusid
import pandas as pd
from lusidtools import cocoon
from lusidtools.cocoon.utilities import checkargs
from lusidtools.cocoon.validator import Validator
from lusidtools.cocoon.async_tools import run_in_executor
import asyncio
from lusidtools.cocoon.dateorcutlabel import DateOrCutLabel
import logging


class BatchLoader:
    """
    This class contains all the methods used for loading data in batches. The @run_in_executor decorator makes the
    synchronous functions awaitable
    """

    @staticmethod
    @run_in_executor
    def load_instrument_batch(api_factory: lusid.utilities.ApiClientFactory,
                              instrument_batch: list, **kwargs) -> lusid.models.UpsertInstrumentsResponse:
        """
        Upserts a batch of instruments to LUSID

        :param lusid.utilities.ApiClientFactory api_factory: The api factory to use
        :param list[lusid.models.InstrumentDefinition] instrument_batch: The batch of instruments to upsert

        :return: UpsertInstrumentsResponse: The response from LUSID
        """

        # Ensure that the list of allowed unique identifiers exists
        if "unique_identifiers" not in list(kwargs.keys()):
            unique_identifiers = cocoon.instruments.get_unique_identifiers(api_factory=api_factory)
        else:
            unique_identifiers = kwargs["unique_identifiers"]

        @checkargs
        def get_alphabetically_first_identifier_key(instrument: lusid.models.InstrumentDefinition,
                                                    unique_identifiers: list):
            """
            Gets the alphabetically first occurring unique identifier on an instrument and use it as the correlation
            id on the request

            :param lusid.models.InstrumentDefinition instrument: The instrument to create a correlation id for
            :param list[str] unique_identifiers: The list of allowed unique identifiers

            :return: str: The correlation id to use on the request
            """
            unique_identifiers_populated = list(set(unique_identifiers).intersection(set(list(instrument.identifiers.keys()))))
            unique_identifiers_populated.sort()
            first_unique_identifier_alphabetically = unique_identifiers_populated[0]
            return f"{first_unique_identifier_alphabetically}: {instrument.identifiers[first_unique_identifier_alphabetically].value}"

        return api_factory.build(lusid.api.InstrumentsApi).upsert_instruments(
            instruments={
                get_alphabetically_first_identifier_key(instrument, unique_identifiers): instrument for instrument in instrument_batch
            })

    @staticmethod
    @run_in_executor
    def load_quote_batch(api_factory: lusid.utilities.ApiClientFactory, quote_batch: list,
                         **kwargs) -> lusid.models.UpsertQuotesResponse:
        """
        Upserts a batch of quotes into LUSID

        :param lusid.utilities.ApiClientFactory api_factory: The api factory to use
        :param list[lusid.models.UpsertQuoteRequest] quote_batch: The batch of quotes to upsert
        :param str scope: The scope to upsert the quotes into

        :return: lusid.models.UpsertQuotesResponse: The response from LUSID
        """

        if "scope" not in list(kwargs.keys()):
            raise KeyError("You are trying to load quotes without a scope, please ensure that a scope is provided.")

        return api_factory.build(lusid.api.QuotesApi).upsert_quotes(
            scope=kwargs["scope"],
            quotes={
                "_".join(
                    [
                        quote.quote_id.quote_series_id.instrument_id,
                        quote.quote_id.quote_series_id.instrument_id_type,
                        str(quote.quote_id.effective_at)
                    ]
                ): quote for quote in quote_batch
            }
        )

    @staticmethod
    @run_in_executor
    def load_transaction_batch(api_factory: lusid.utilities.ApiClientFactory, transaction_batch: list,
                               **kwargs) -> lusid.models.UpsertPortfolioTransactionsResponse:
        """
        Upserts a batch of transactions into LUSID

        :param lusid.utilities.ApiClientFactory api_factory: The api factory to use
        :param str scope: The scope of the Transaction Portfolio to upsert the transactions into
        :param str code: The code of the Transaction Portfolio, together with the scope this uniquely identifies the portfolio
        :param list[lusid.models.TransactionRequest] transaction_batch: The batch of transactions to upsert

        :return: lusid.models.UpsertPortfolioTransactionsResponse: The response from LUSID
        """

        if "scope" not in list(kwargs.keys()):
            raise KeyError("You are trying to load transactions without a scope, please ensure that a scope is provided.")

        if "code" not in list(kwargs.keys()):
            raise KeyError("You are trying to load transactions without a portfolio code, please ensure that a code is provided.")

        return api_factory.build(lusid.api.TransactionPortfoliosApi).upsert_transactions(
            scope=kwargs["scope"],
            code=kwargs["code"],
            transactions=transaction_batch)

    @staticmethod
    @run_in_executor
    def load_holding_batch(api_factory: lusid.utilities.ApiClientFactory, holding_batch: list,
                           **kwargs) -> lusid.models.HoldingsAdjustment:
        """
        Upserts a batch of holdings into LUSID

        :param lusid.utilities.ApiClientFactory api_factory: The api factory to use
        :param list[lusid.models.AdjustHoldingRequest] holding_batch: The batch of holdings
        :param kwargs: 'scope', 'code', 'effective_at' The parameters required for the API call

        :return: lusid.models.HoldingsAdjustment: The response from LUSID
        """

        if "scope" not in list(kwargs.keys()):
            raise KeyError("You are trying to load transactions without a scope, please ensure that a scope is provided.")

        if "code" not in list(kwargs.keys()):
            raise KeyError("You are trying to load transactions without a portfolio code, please ensure that a code is provided.")

        if "effective_at" not in list(kwargs.keys()):
            raise KeyError('''There is no mapping for effective_at in the required mapping, please add it''')

        return api_factory.build(lusid.api.TransactionPortfoliosApi).set_holdings(
            scope=kwargs["scope"],
            code=kwargs["code"],
            effective_at=str(DateOrCutLabel(kwargs["effective_at"])),
            holding_adjustments=holding_batch)

    @staticmethod
    @run_in_executor
    def load_portfolio_batch(api_factory: lusid.utilities.ApiClientFactory, portfolio_batch: list,
                             **kwargs) -> lusid.models.Portfolio:
        """
        Upserts a batch of portfolios to LUSID

        :param lusid.utilities.ApiClientFactory api_factory: The api factory to use
        :param list[lusid.models.CreateTransactionPortfolioRequest] portfolio_batch: The batch of portfolios to create
        :param kwargs: 'scope', 'code' arguments required for the API call

        :return: lusid.models.Portfolio: The response from LUSID
        """

        if "scope" not in list(kwargs.keys()):
            raise KeyError("You are trying to load transactions without a scope, please ensure that a scope is provided.")

        if "code" not in list(kwargs.keys()):
            raise KeyError("You are trying to load transactions without a portfolio code, please ensure that a code is provided.")

        try:
            return api_factory.build(lusid.api.PortfoliosApi).get_portfolio(
                scope=kwargs["scope"],
                code=kwargs["code"]
            )
        except lusid.exceptions.ApiException as e:
            if e.status == 404:
                return api_factory.build(lusid.api.TransactionPortfoliosApi).create_portfolio(
                    scope=kwargs["scope"],
                    transaction_portfolio=portfolio_batch[0])
            # Add in here upsert portfolio properties if it does exist


async def load_data(api_factory: lusid.utilities.ApiClientFactory, data_frame: pd.DataFrame,
                    mapping_required: dict, mapping_optional: dict,
                    property_columns: list, properties_scope: str, instrument_identifier_mapping: dict,
                    file_type: str, domain_lookup: dict, **kwargs):
    """
    This function populates the required models from a DataFrame and loads the data into LUSID

    :param lusid.utilities.ApiClientFactory api_factory: The api factory to use
    :param pd.DataFrame data_frame: The DataFrame containing the data to load
    :param dict mapping_required: The required mapping
    :param dict mapping_optional: The optional mapping
    :param list property_columns: The property columns to add as property values
    :param str properties_scope: The scope to add the property values in
    :param dict instrument_identifier_mapping: The mapping for the identifiers
    :param int batch_size: The batch size to use
    :param str file_type: The file type to load
    :param dict domain_lookup: The domain lookup
    :param kwargs: Arguments specific to each call e.g. effective_at for holdings

    :return: BatchLoader StaticMethod: A static method on BatchLoader
    """

    # Get the top level model used for this request e.g. lusid.models.InstrumentDefintion for upsert instruments
    top_level_model = getattr(lusid.models, domain_lookup[file_type]["top_level_model"])

    # Verify that all the required attributes for this top level model exist in the provided required mapping
    cocoon.utilities.verify_all_required_attributes_mapped(
        swagger_dict=cocoon.utilities.get_swagger_dict(api_url=api_factory.api_client.configuration.host),
        mapping=mapping_required,
        model_object=top_level_model,
        exempt_attributes=["identifiers", "properties", "instrument_identifiers"])

    # Get the data types of the columns to be added as properties
    property_dtypes = data_frame.loc[:, property_columns].dtypes

    unique_identifiers = kwargs['unique_identifiers']

    # Iterate over the DataFrame creating the single requests
    single_requests = []
    for index, row in data_frame.iterrows():

        # Create the property values for this row
        if domain_lookup[file_type]['domain'] is None:
            properties = None
        else:
            properties = cocoon.properties.create_property_values(
                row=row,
                scope=properties_scope,
                domain=domain_lookup[file_type]['domain'],
                dtypes=property_dtypes)

        # Create identifiers for this row if applicable
        # If no instrument identifier mapping is provided return None as no identifiers are required
        if instrument_identifier_mapping is None or not bool(instrument_identifier_mapping):
            identifiers = None
        else:
            identifiers = cocoon.instruments.create_identifiers(
                index=index,
                row=row,
                file_type=file_type,
                instrument_identifier_mapping=instrument_identifier_mapping,
                unique_identifiers=unique_identifiers,
                full_key_format=kwargs["full_key_format"],
            )

        # Construct the from the mapping, properties and identifiers the single request object and add it to the list
        single_requests.append(cocoon.utilities.populate_model(
            model_object=top_level_model,
            required_mapping=mapping_required,
            optional_mapping=mapping_optional,
            row=row,
            properties=properties,
            identifiers=identifiers))

    # Dynamically call the correct async function to use based on the file type
    return await getattr(BatchLoader, f"load_{file_type}_batch")(
        api_factory,
        single_requests,
        # Any specific arguments e.g. 'code' for transactions, 'effective_at' for holdings is passed in via **kwargs
        **kwargs
    )


async def construct_batches(api_factory: lusid.utilities.ApiClientFactory, data_frame: pd.DataFrame,
                            mapping_required: dict, mapping_optional: dict,
                            property_columns: list, properties_scope: str, instrument_identifier_mapping: dict,
                            batch_size: int, file_type: str, domain_lookup: dict, **kwargs):

    """
    This constructs the batches and asynchronously sends them to be loaded into LUSID

    :param lusid.utilities.ApiClientFactory api_factory: The api factory to use
    :param pd.DataFrame data_frame: The DataFrame containing the data to load
    :param dict mapping_required: The required mapping
    :param dict mapping_optional: The optional mapping
    :param list property_columns: The property columns to add as property values
    :param str properties_scope: The scope to add the property values in
    :param dict instrument_identifier_mapping: The mapping for the identifiers
    :param int batch_size: The batch size to use
    :param str file_type: The file type to load
    :param dict domain_lookup: The domain lookup
    :param kwargs: Arguments specific to each call e.g. effective_at for holdings

    :return: dict: Contains the success responses and the errors (where an API exception has been raised)
    """

    # Get the different behaviours required for different entities e.g quotes can be batched without worrying about portfolios
    batching_no_portfolios = [file_type for file_type, settings in domain_lookup.items() if not settings["portfolio_specific"]]
    batching_with_portfolios = [file_type for file_type, settings in domain_lookup.items() if settings["portfolio_specific"]]

    if file_type in batching_no_portfolios:

        async_batches = [data_frame.iloc[i:i+batch_size] for i in range(0, len(data_frame), batch_size)]

        sync_batches = [
            {
                "async_batches": async_batches,
                "codes": [None] * len(async_batches),
                "effective_at": [None] * len(async_batches)
            }
        ]

    elif file_type in batching_with_portfolios:

        if 'effective_at' in domain_lookup[file_type]["required_call_attributes"]:

            unique_effective_dates = list(data_frame[mapping_required['effective_at']].unique())

            effective_at_groups = [
                data_frame.loc[
                    data_frame[mapping_required['effective_at']] == effective_at] for effective_at in unique_effective_dates
            ]

            sync_batches = [
                {
                    "async_batches": [
                        data_frame.loc[
                            data_frame[mapping_required['code']] == code] for code in list(effective_at_group[mapping_required['code']].unique())
                    ],
                    "codes": list(effective_at_group[mapping_required['code']].unique()),
                    "effective_at": [list(effective_at_group[mapping_required['effective_at']].unique())[0]] * len(list(effective_at_group[mapping_required['code']].unique()))
                } for effective_at_group in effective_at_groups
            ]

        else:

            unique_portfolios = list(data_frame[mapping_required['code']].unique())

            async_batches = [
                data_frame.loc[
                    data_frame[mapping_required['code']] == code] for code in unique_portfolios
            ]

            sync_batches = [
                {
                    "async_batches": async_batches,
                    "codes": [str(code) for code in unique_portfolios],
                    "effective_at": [None] * len(async_batches)
                }
            ]


    # Asynchronously load the data into LUSID
    responses = [await asyncio.gather(
        *[load_data(
            api_factory=api_factory,
            data_frame=async_batch,
            mapping_required=mapping_required,
            mapping_optional=mapping_optional,
            property_columns=property_columns,
            properties_scope=properties_scope,
            instrument_identifier_mapping=instrument_identifier_mapping,
            file_type=file_type,
            domain_lookup=domain_lookup,
            code=code,
            effective_at=effective_at,
            **kwargs
        ) for async_batch, code, effective_at in zip(sync_batch["async_batches"], sync_batch["codes"], sync_batch["effective_at"])],
        return_exceptions=True
    ) for sync_batch in sync_batches]

    responses_flattened = [response for responses_sub in responses for response in responses_sub]

    # Raise any internal exceptions rather than propagating them to the response
    for response in responses_flattened:
        if isinstance(response, Exception) and not isinstance(response, lusid.exceptions.ApiException):
            raise response

    # Collects the exceptions as failures and successful calls as values
    return {
        "errors": [r for r in responses_flattened if isinstance(r, Exception)],
        "success": [r for r in responses_flattened if not isinstance(r, Exception)]
    }


@checkargs
def load_from_data_frame(api_factory: lusid.utilities.ApiClientFactory, scope: str, data_frame: pd.DataFrame,
                         mapping_required: dict, mapping_optional: dict, file_type: str,
                         identifier_mapping: dict = None, property_columns: list = None,
                         properties_scope: str = None, batch_size: int = None, instrument_name_enrichment: bool = False):
    """
    Handles loading data from a DataFrame into LUSID

    :param lusid.utilities.ApiClientFactory api_factory: The api factory to use
    :param str scope: The scope of the resource to load the data into
    :param Pandas DataFrame data_frame: The DataFrame containing the data
    :param dict{str, str} mapping_required: The dictionary mapping the DataFrame columns to LUSID's required attributes
    :param dict{str, str} mapping_optional: The dictionary mapping the DataFrame columns to LUSID's optional attributes
    :param str file_type: The type of file e.g. transctions, instruments, holdings, quotes, portfolios
    :param dict{str, str} identifier_mapping: The dictionary mapping of LUSID instrument identifiers to identifiers in the DataFrame
    :param list[str] property_columns: The columns to create properties for
    :param str properties_scope: The scope to add the properties to
    :param int batch_size: The size of the batch to use when using upsert calls e.g. upsert instruments, upsert quotes etc.

    :return: dict responses: The responses from loading the data into LUSID
    """

    # A mapping between the file type and relevant attributes e.g. domain, top_level_model etc.
    domain_lookup = cocoon.utilities.load_json_file('config/domain_settings.json')

    # Convert the file type to lower case & singular as well as checking it is of the allowed value
    file_type = Validator(file_type, 'file_type')\
        .make_singular()\
        .make_lower()\
        .check_allowed_value(list(domain_lookup.keys()))\
        .value

    # Set defaults aligned with the data type of each argument, this allows for users to provide None
    identifier_mapping = Validator(identifier_mapping, 'identifier_mapping')\
        .set_default_value_if_none(default={})\
        .discard_dict_keys_none_value()\
        .value

    properties_scope = Validator(properties_scope, 'properties_scope')\
        .set_default_value_if_none(default=scope)\
        .value

    property_columns = Validator(property_columns, 'property_columns')\
        .set_default_value_if_none(default=[])\
        .value

    batch_size = Validator(batch_size, 'batch_size')\
        .set_default_value_if_none(domain_lookup[file_type]["default_batch_size"])\
        .override_value(not domain_lookup[file_type]["batch_allowed"], domain_lookup[file_type]["default_batch_size"])\
        .value

    # Discard mappings where the provided value is None
    mapping_required = Validator(mapping_required, 'mapping_required')\
        .discard_dict_keys_none_value()\
        .value

    mapping_optional = Validator(mapping_optional, 'mapping_optional') \
        .discard_dict_keys_none_value() \
        .value

    if instrument_name_enrichment:

        loop = cocoon.async_tools.start_event_loop_new_thread()

        data_frame, mapping_required = asyncio.run_coroutine_threadsafe(
            cocoon.instruments.enrich_instruments(
                api_factory=api_factory,
                data_frame=data_frame,
                instrument_identifier_mapping=identifier_mapping,
                mapping_required=mapping_required,
                constant_prefix="$"),
            loop
        ).result()

        # Stop the additional event loop
        loop.stop()

    '''
    Unnest and populate defaults where a mapping is provided with column and/or default fields in a nested dictionary
    
    e.g.
    {'name': {
        'column': 'instrument_name',
        'default': 'unknown_name'
        }
    }
    
    rather than simply
    {'name': 'instrument_name'}
    '''
    data_frame, mapping_required = cocoon.utilities.handle_nested_default_and_column_mapping(
        data_frame=data_frame,
        mapping=mapping_required,
        constant_prefix="$"
    )
    data_frame, mapping_optional = cocoon.utilities.handle_nested_default_and_column_mapping(
        data_frame=data_frame,
        mapping=mapping_optional,
        constant_prefix="$"
    )

    # Get all the DataFrame columns as well as those that contain at least one null value
    data_frame_columns = list(data_frame.columns.values)
    nan_columns = [column for column in data_frame_columns if data_frame[column].isna().any()]

    # Validate that none of the provided columns are missing or invalid
    Validator(mapping_required, 'mapping_required')\
        .get_dict_values()\
        .filter_list_using_first_character("$")\
        .check_subset_of_list(data_frame_columns, "DataFrame Columns")\
        .check_no_intersection_with_list(nan_columns, 'Columns with Missing Values')

    Validator(mapping_optional, 'mapping_optional')\
        .get_dict_values()\
        .filter_list_using_first_character("$")\
        .check_subset_of_list(data_frame_columns, "DataFrame Columns")

    Validator(identifier_mapping, 'identifier_mapping')\
        .get_dict_values()\
        .filter_list_using_first_character("$")\
        .check_subset_of_list(data_frame_columns, "DataFrame Columns")

    Validator(property_columns, 'property_columns')\
        .check_subset_of_list(data_frame_columns, "DataFrame Columns")

    # Converts higher level data types such as dictionaries and lists to strings
    data_frame = data_frame.applymap(cocoon.utilities.convert_cell_value_to_string)

    # Check for and create missing property defintions
    data_frame = cocoon.properties.create_missing_property_definitions_from_file(
        api_factory=api_factory,
        properties_scope=properties_scope,
        file_type=file_type,
        data_frame=data_frame,
        property_columns=property_columns,
        domain_lookup=domain_lookup)

    # Start a new event loop in a new thread, this is required to run inside a Jupyter notebook
    loop = cocoon.async_tools.start_event_loop_new_thread()

    # Keyword arguments to be used in requests to the LUSID API
    keyword_arguments = {
        "scope": scope,
        # This handles that identifiers need to be specified differently based on the request type, allowing users
        # to provide either the entire key e.g. "Instrument/default/Figi" or just the code "Figi" for any request
        "full_key_format": domain_lookup[file_type]["full_key_format"],
        # Gets the allowed unique identifiers
        "unique_identifiers": cocoon.instruments.get_unique_identifiers(api_factory=api_factory)
    }

    # Get the responses from LUSID
    responses = asyncio.run_coroutine_threadsafe(
        construct_batches(
            api_factory=api_factory,
            data_frame=data_frame,
            mapping_required=mapping_required,
            mapping_optional=mapping_optional,
            property_columns=property_columns,
            properties_scope=properties_scope,
            instrument_identifier_mapping=identifier_mapping,
            batch_size=batch_size,
            file_type=file_type,
            domain_lookup=domain_lookup,
            **keyword_arguments),
        loop
        ).result()

    # Stop the additional event loop
    loop.stop()

    # Prefix the responses with the file type
    return {file_type+'s': responses}



import lusid
import lusid.models as models
from lusid.api import InstrumentsApi, SearchApi
import numpy as np
import time
from lusidtools.cocoon.utilities import checkargs
import pandas as pd
import logging
import re
from lusidtools.cocoon.async_tools import run_in_executor
import asyncio


@checkargs
def create_identifiers(index: int, row: pd.Series, file_type: str, instrument_identifier_mapping: dict = None,
                       unique_identifiers: list = None, full_key_format: bool = True):
    """
    This function creates identifiers to be added to a LUSID model object

    :param int index: The index of the row in the DataFrame
    :param pd.Series row: The row of the DataFrame to create identifiers for
    :param str file_type: The file type to create identifiers for
    :param dict instrument_identifier_mapping: The instrument identifier mapping to use
    :param list unique_identifiers: The list of allowable unique instrument identifiers
    :param bool full_key_format: Whether or not the full key format i.e. 'Instrument/default/Figi' is required

    :return: dict identifiers: The identifiers to use on the request
    """

    @checkargs
    def prepare_key(identifier_lusid: str, full_key_format: bool) -> str:
        """
        This function prepares the key for the identifier based on whether the full key or just the code is required

        :param str identifier_lusid: The LUSID identifier in either full key format or code only
        :param bool full_key_format: Whether or not they key needs to be in the full key format

        :return: str: The LUSID identifier in the correct format
        """
        if full_key_format:
            return identifier_lusid if re.findall('Instrument/default/\S+', identifier_lusid) else f'Instrument/default/{identifier_lusid}'
        else:
            return identifier_lusid.split("/")[2] if re.findall('Instrument/default/\S+', identifier_lusid) else identifier_lusid

    # Populate the identifiers for this instrument
    identifiers = {
        # Handles specifying the entire property key e.g. Instrument/default/Figi or just the code e.g. Figi
        prepare_key(identifier_lusid, full_key_format):
            models.InstrumentIdValue(value=row[identifier_column]) if file_type == 'instrument' else row[identifier_column] for \
        identifier_lusid, identifier_column in instrument_identifier_mapping.items() if not \
            # Only use the identifier it it has a value
        pd.isna(row[identifier_column])
    }

    # If there are no identifiers raise an error
    if len(identifiers) == 0:
        raise ValueError(f'''The row at index {index} has no value for every single one of the provided 
        identifiers. Please ensure that each row has at least one identifier and try again''')

    # Check that at least one unique identifier exists if it is an instrument file (need to move this out of here)
    if file_type == 'instrument':

        # Get the unique list of unique identifiers which have been populated
        unique_identifiers_populated = list(set(unique_identifiers).intersection(set(list(identifiers.keys()))))

        # If there are no unique identifiers raise an Exception as you need at least one to make a successful call
        if len(unique_identifiers_populated) == 0:
            raise ValueError(f'''The instrument at index {index} has no value for at least one unique 
            identifier. Please ensure that each instrument has at least one unique identifier and try again. The
            allowed unique identifiers are {str(unique_identifiers)}''')

    else:
        # If the transaction/holding is cash remove all other identifiers and just use this one
        if "Instrument/default/Currency" in list(identifiers.keys()):
            currency_value = identifiers["Instrument/default/Currency"]
            identifiers.clear()
            identifiers["Instrument/default/Currency"] = currency_value

    return identifiers


@checkargs
def resolve_instruments(api_factory: lusid.utilities.ApiClientFactory, data_frame: pd.DataFrame, identifier_mapping: dict):
    """
    This function attempts to resolve each row of the file to an instrument in LUSID

    :param lusid.utilities.ApiClientFactory api_factory: An instance of the Lusid Api Factory
    :param Pandas DataFrame data_frame: The DataFrame containing the transactions or holdings to resolve to unique instruments
    :param dict identifier_mapping: The column mapping between allowable identifiers in LUSID
                                    and identifier columns in the dataframe

    :return: Pandas DataFrame _data_frame: The input DataFrame with resolution columns added
    """

    if 'Currency' not in list(identifier_mapping.keys()) and 'Instrument/default/Currency' not in list(identifier_mapping.keys()):
        raise KeyError(
            '''There is no column specified in the identifier_mapping to identify whether or not an instrument is cash.
               Please specify add the key "Currency" or "Instrument/default/Currency" to your mapping. If no instruments
               are cash you can set the value to be None''')

    if 'Currency' in list(identifier_mapping.keys()):
        identifier_mapping['Instrument/default/Currency'] = identifier_mapping['Currency']
        del identifier_mapping['Currency']

    # Check that the values of the mapping exist in the DataFrame
    if not (set(identifier_mapping.values()) <= set(data_frame.columns)):
        raise KeyError('there are values in identifier_mapping that are not columns in the dataframe')

    # Get the allowable instrument identifiers from LUSID
    response = api_factory.build(InstrumentsApi).get_instrument_identifier_types()
    """
    # Collect the names and property keys for the identifiers and concatenate them
    allowable_identifier_names = [identifier.identifier_type for identifier in response.values]
    allowable_identifier_keys = [identifier.property_key for identifier in response.values]
    allowable_identifiers = allowable_identifier_names + allowable_identifier_keys

    # Check that the identifiers in the mapping are all allowed to be used in LUSID
    if not (set(identifier_mapping['identifier_mapping'].keys()) <= set(allowable_identifiers)):
        raise Exception(
            'there are LUSID identifiers in the identifier_mapping which are not configured in LUSID')
    """
    # Copy the data_frame to ensure the original isn't modified
    _data_frame = data_frame.copy(deep=True)

    # Set up the result Pandas Series to track resolution
    found_with = pd.Series(index=_data_frame.index, dtype=np.dtype(object))
    resolvable = pd.Series(index=_data_frame.index, dtype=np.dtype(bool))
    luid = pd.Series(index=_data_frame.index, dtype=np.dtype(object))
    comment = pd.Series(index=_data_frame.index, dtype=np.dtype(object))
    logging.info('Beginning instrument resolution process')
    # Iterate over each row in the DataFrame
    for index, row in _data_frame.iterrows():

        if index % 10 == 0:
            logging.info(f'Up to row {index}')
        # Initialise list to hold the identifiers used to resolve
        found_with_current = []
        # Initialise a value of False for the row's resolvability to an instrument in LUSID
        resolvable_current = False
        # Initilise the LUID value
        luid_current = None
        # Initialise the comment value
        comment_current = "No instruments found for the given identifiers"
        # Takes the currency resolution function and applies it
        currency = row[identifier_mapping['Instrument/default/Currency']]

        if not pd.isna(currency):
            resolvable_current = True
            found_with_current.append(currency)
            luid_current = currency
            comment_current = 'Resolved as cash with a currency'

        search_requests = [
            models.InstrumentSearchProperty(
                key=f'Instrument/default/{identifier_lusid}' if 'Instrument/' not in identifier_lusid else identifier_lusid,
                value=row[identifier_dataframe]
            ) for identifier_lusid, identifier_dataframe in identifier_mapping.items() if not pd.isnull(row[identifier_dataframe])
        ]

        # Call LUSID to search for instruments
        attempts = 0

        if len(search_requests) > 0:
            while attempts < 3:
                try:
                    response = api_factory.build(SearchApi).instruments_search(
                        symbols=search_requests,
                        mastered_only=True)
                    break
                except lusid.exceptions.ApiException as error_message:
                    attempts += 1
                    comment_current = f'Failed to find instrument due to LUSID error during search due to status {error_message.status} with reason {error_message.reason}'
                    time.sleep(5)

            if attempts == 3:
                # Update the luid series
                luid.iloc[index] = luid_current
                # Update the found with series
                found_with.iloc[index] = found_with_current
                # Update the resolvable series
                resolvable.iloc[index] = resolvable_current
                # Update the comment series
                comment.iloc[index] = comment_current
                continue

            search_request_number = -1

            for result in response:

                search_request_number += 1
                # If there are matches
                if len(result.mastered_instruments) == 1:
                    # Add the identifier responsible for the successful search request to the list
                    found_with_current.append(
                        search_requests[
                            search_request_number].key.split('/')[2]
                    )
                    comment_current = 'Uniquely resolved to an instrument in the securities master'
                    resolvable_current = True
                    luid_current = result.mastered_instruments[0].identifiers['LusidInstrumentId'].value
                    break

                elif len(result.mastered_instruments) > 1:
                    comment_current = f'Multiple instruments found for the instrument using identifier {search_requests[search_request_number].key.split("/")[2]}'
                    resolvable_current = False
                    luid_current = np.NaN

        # Update the luid series
        luid.iloc[index] = luid_current
        # Update the found with series
        found_with.iloc[index] = found_with_current
        # Update the resolvable series
        resolvable.iloc[index] = resolvable_current
        # Update the comment series
        comment.iloc[index] = comment_current

    # Add the series to the dataframe
    _data_frame['resolvable'] = resolvable
    _data_frame['foundWith'] = found_with
    _data_frame['LusidInstrumentId'] = luid
    _data_frame['comment'] = comment

    return _data_frame


@checkargs
def get_unique_identifiers(api_factory: lusid.utilities.ApiClientFactory):
    """
    Tests getting the unique instrument identifiers

    :param lusid.utilities.ApiClientFactory api_factory: The LUSID api factory to use
    :return: list[str]: The property keys of the available identifiers
    """

    # Get the allowed instrument identifiers from LUSID
    identifiers = api_factory.build(lusid.api.InstrumentsApi).get_instrument_identifier_types()

    # Return the identifiers that are configured to be unique
    return [identifier.identifier_type for identifier in identifiers.values if identifier.is_unique_identifier_type]


@checkargs
async def enrich_instruments(api_factory: lusid.utilities.ApiClientFactory, data_frame: pd.DataFrame,
                             instrument_identifier_mapping: dict, mapping_required: dict, constant_prefix: str = "$"):

    search_requests_all = []

    for index, row in data_frame.iterrows():

        search_requests_instrument = [
            lusid.models.InstrumentSearchProperty(
                key=identifier_lusid if re.findall('Instrument/default/\S+', identifier_lusid) else f'Instrument/default/{identifier_lusid}',
                value=row[identifier_column]
            ) for identifier_lusid, identifier_column in instrument_identifier_mapping.items() if not pd.isna(row[identifier_column])
        ]

        search_requests_all.append(search_requests_instrument)

    responses = await asyncio.gather(
        *[instrument_search(
            api_factory=api_factory,
            search_requests=search_requests
        ) for search_requests in search_requests_all],
        return_exceptions=False
    )

    names = []

    for response in responses:
        name = np.NaN
        for single_search in response:
            if isinstance(single_search, Exception):
                logging.warning(single_search)
                continue
            elif len(single_search[0].external_instruments) > 0:
                name = single_search[0].external_instruments[0].name
                break
        names.append(name)

    enriched_column_name = 'LUSID.Name.Enriched'

    data_frame[enriched_column_name] = names

    # Missing mapping for name altogether
    if "name" not in list(mapping_required.keys()):
        mapping_required["name"] = enriched_column_name

    # A column for name already exists and needs to be enriched
    elif isinstance(mapping_required["name"], str) and mapping_required["name"][0] != constant_prefix:
        data_frame[mapping_required["name"]] = data_frame[mapping_required["name"]].fillna(
            value=data_frame[enriched_column_name])

    elif isinstance(mapping_required["name"], dict) and "column" in list(mapping_required["name"].keys()):
        data_frame[mapping_required["name"]["column"]] = data_frame[mapping_required["name"]["column"]].fillna(
            value=data_frame[enriched_column_name])

    # Is a constant specified by the constant prefix
    elif isinstance(mapping_required["name"], str) and mapping_required["name"][0] == constant_prefix:
        mapping_required["name"] = {
            "default": mapping_required["name"][1:]
        }
        mapping_required["name"]["column"] = enriched_column_name

    # Is a constant specified by the default nested dictionary specification
    elif isinstance(mapping_required["name"], dict) and "default" in list(mapping_required["name"].keys()) and "column" not in list(mapping_required["name"].keys()):
        mapping_required["name"]["column"] = enriched_column_name

    return data_frame, mapping_required


async def instrument_search(api_factory: lusid.utilities.ApiClientFactory, search_requests: list) -> list:
    """
    Conducts an instrument search for a single instrument

    :param lusid.utilities.ApiClientFactory api_factory: The api factory to use
    :param list[lusid.models.InstrumentSearchProperty] search_requests: The search requests for this instrument

    :return:
    """

    instrument_search_results = []

    for search_request in search_requests:
        try:
            result = await instrument_search_single(api_factory, search_request)
            instrument_search_results.append(result)
        except lusid.exceptions.ApiException as e:
            instrument_search_results.append(e)
    return instrument_search_results


@run_in_executor
def instrument_search_single(api_factory: lusid.utilities.ApiClientFactory,
                             search_request: lusid.models.InstrumentSearchProperty) -> list:
    """
    Conducts an instrument search with a single search request

    :param lusid.utilities.ApiClientFactory api_factory: The api factory to use
    :param lusid.models.InstrumentSearchProperty search_request: The search request

    :return: list[lusid.models.InstrumentMatch]: The results of the search
    """

    return lusid.api.SearchApi(api_factory.build(lusid.api.SearchApi)).instruments_search(
        symbols=[search_request]
    )



import json
import logging
import os
import sys
from lusid.utilities import ApiClientFactory
from lusidtools.cocoon import load_from_data_frame, load_data_to_df_and_detect_delimiter, \
    check_mapping_fields_exist, parse_args, validate_mapping_file_structure
from lusidtools.logger import LusidLogger


def load_mapping_file_for_file_type(mapping_path, file_type) -> dict:
    """
    :param mapping_path: The full path of mapping_file.json
    :param file_type: The type of file i.e. transactions or holdings or instruments
    :return:
    """
    if not os.path.exists(mapping_path):
        raise OSError(f"mapping file not found at {mapping_path}")
    with open(mapping_path, "r") as read_file:
        return json.load(read_file)[file_type]


def load_instruments(args):
    file_type = "instruments"

    # create ApiFactory
    factory = ApiClientFactory(api_secrets_filename=args["secrets_file"])

    # get data
    if args['delimiter']:
        logging.info(f"delimiter specified as {repr(args['delimiter'])}")
    logging.debug("Getting data")
    instruments = load_data_to_df_and_detect_delimiter(args)

    # get mappings
    mappings = load_mapping_file_for_file_type(args["mapping"], file_type)

    if "property_columns" not in mappings.keys() and not args["scope"]:
        raise ValueError(r"Instrument properties must be upserted to a specified scope, but no scope was provided. "
                         r"Please state what scope to upsert properties to using '-s'.")

    validate_mapping_file_structure(mappings, instruments.columns, file_type)

    if args["dryrun"]:
        logging.info("--dryrun specified as True, exiting before upsert call is made")
        return 0

    instruments_response = load_from_data_frame(api_factory=factory,
                                                data_frame=instruments,
                                                scope=args["scope"],
                                                mapping_required=mappings["required"],
                                                mapping_optional=mappings["optional"],
                                                file_type=file_type,
                                                identifier_mapping=mappings["identifier_mapping"],
                                                batch_size=args["batch_size"],
                                                property_columns=mappings[
                                                    "property_columns"] if "property_columns" in mappings.keys() else [])

    total_success = sum([len(resp.values.keys()) for resp in instruments_response['instruments']['success']])
    total_failed = sum([len(resp.failed.keys()) for resp in instruments_response['instruments']['success']])
    logging.info(f"Success: {total_success}/{len(instruments)}")
    logging.info(f"Fail:    {total_failed}/{len(instruments)}")
    return instruments_response


def main(argv):
    args, ap = parse_args(sys.argv[1:])
    LusidLogger(args["debug"])
    load_instruments(args)

    return 0


if __name__ == "__main__":
    main(sys.argv)



import sys
import logging
from lusid.utilities import ApiClientFactory

from lusidtools.apps.upsert_instruments import load_mapping_file_for_file_type
from lusidtools.cocoon import load_data_to_df_and_detect_delimiter, load_from_data_frame, parse_args, \
    identify_cash_items, validate_mapping_file_structure
from lusidtools.logger import LusidLogger


def load_transactions(args):
    file_type = "transactions"

    factory = ApiClientFactory(api_secrets_filename=args["secrets_file"])

    if args['delimiter']:
        logging.info(f"delimiter specified as {repr(args['delimiter'])}")
    logging.debug("Getting data")
    transactions = load_data_to_df_and_detect_delimiter(args)

    mappings = load_mapping_file_for_file_type(args["mapping"], file_type)

    if "cash_flag" in mappings.keys():
        identify_cash_items(transactions, mappings)

    validate_mapping_file_structure(mappings, transactions.columns, file_type)

    if args["dryrun"]:
        logging.info("--dryrun specified as True, exiting before upsert call is made")
        return 0

    transactions_response = load_from_data_frame(api_factory=factory,
                                                 data_frame=transactions,
                                                 scope=args["scope"],
                                                 identifier_mapping=mappings["identifier_mapping"],
                                                 mapping_required=mappings["required"],
                                                 mapping_optional=mappings["optional"],
                                                 file_type=file_type,
                                                 batch_size=args["batch_size"],
                                                 property_columns=mappings[
                                                     "property_columns"] if "property_columns" in mappings.keys() else [])

    # print_response(transactions_response, file_type)
    total_success = len(transactions_response["transactions"]["success"])
    total_failed = len(transactions_response["transactions"]["errors"])
    logging.info(f"Success: {total_success}/{total_success + total_failed}")
    logging.info(f"Fail:    {total_failed}/{total_success + total_failed}")
    return transactions_response


def main(argv):
    args, ap = parse_args(sys.argv[1:])
    LusidLogger(args["debug"])
    load_transactions(args)

    return 0


if __name__ == "__main__":
    main(sys.argv)
